#include <cmath>
#include <cstring>

#include <iostream>

#include "stm.h"

#include <pthread.h>
#include "../common/time.h"
#include "thread.h"
#include "thread_fun.h"

#include "../parameters.h"
#include "../data_holder.h"
#include "../sb7_exception.h"
#include "../helpers.h"

void *sb7::init_data_holder(void *data) {
	thread_init();

	BEGIN_TRANSACTION;

	DataHolder *dataHolder = (DataHolder *)data;
	dataHolder->init();

	END_TRANSACTION;

	// just return something
	return NULL;
}

void *sb7::worker_thread(void *data) {
	thread_init();

	WorkerThreadData *wtdata = (WorkerThreadData *)data;

	while(!wtdata->stopped) {
		int opind = wtdata->getOperationRndInd();
		const Operation *op = wtdata->operations->getOperations()[opind];

		try {
			// get start time
			long long start_time = get_time_ns();

			// execute transaction
			while(true) {
				try {
					swisstm::start_tx();

					// transaction body
					op->run();

					swisstm::commit_tx();
					break;
				} catch (swisstm::AbortedException) {
					swisstm::abort_tx();

					// count aborts
					wtdata->aborted_ops[opind]++;
				} catch (...) {
					swisstm::abort_tx_full();
            		throw;
        		}
			}

			// get end time
			long long end_time = get_time_ns();

			wtdata->successful_ops[opind]++;
			long long ttc_ns = end_time - start_time;
			wtdata->total_ttc[opind] += ttc_ns / NANOSECONDS_IN_MICROSECOND;
			int ttc = (int)(ttc_ns / NANOSECONDS_IN_MILLISECOND);

			if(ttc <= wtdata->max_low_ttc) {
				if(ttc < 0) {
					std::cout << "ttc lower than 0 " << std::endl;
				} else {
					wtdata->operations_ttc[opind][ttc]++;
				}
			} else {
				double logHighTtc = (::log(ttc) - wtdata->max_low_ttc_log) /
					wtdata->high_ttc_log_base;
				int intLogHighTtc =
					MIN((int)logHighTtc, wtdata->high_ttc_entries - 1);
				wtdata->operations_high_ttc_log[opind][intLogHighTtc]++;
			}
		} catch (Sb7Exception) {
			wtdata->failed_ops[opind]++;
		}
	}

	thread_clean();

	// just return something
	return NULL;
}

int sb7::WorkerThreadData::getOperationRndInd() const {
	double oprnd = get_random()->nextDouble();
	const std::vector<double> &opRat = operations->getOperationCdf();
	int opind = 0;

	while(opRat[opind] < oprnd) {
		opind++;
	}

	return opind;
}
